/*
 * bgx.c
 * Copyright (C) 2019 Peter Belkner <pbelkner@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#if defined (_WIN32) // [
#include <windows.h>
#include <direct.h>
#else // ] [
#include <ctype.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#endif // ]
#include <bg.h>

///////////////////////////////////////////////////////////////////////////////
static bg_tree_vtbl_t album_vtbl;

int bg_album_create(bg_tree_t *tree, bg_param_t *param)
{
  int err=-1;
  const ffchar_t *path=tree->source.path;
  bg_album_t *album=&tree->album;
  bg_tree_t *child=NULL;
#if defined (_WIN32) // [
  size_t len=wcslen(path);
  wchar_t *mask,*mp;
  HANDLE hFind;
  WIN32_FIND_DATAW e;
#else // ] [
  DIR *dir;
  struct dirent *e;
#endif // ]
  ffchar_t *name;

  /////////////////////////////////////////////////////////////////////////////
  tree->vtbl=&album_vtbl;
  album->first=NULL;
  album->count=0ul;
#if defined (BG_PARAM_SEQUENTIAL) // [
  album->nleaves=0l;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  if (!path) {
    DMESSAGE("attempt to open NULL directory");
    goto edir;
  }

#if defined (_WIN32) // [
  mask=mp=malloc((len+3)*sizeof mask[0]);

  if (!mask) {
    DMESSAGE("allocating directory mask");
    goto emask;
  }

  memcpy(mp,path,len*sizeof mp[0]);
  mp+=len;
  *mp++=FFPATHSEP;
  *mp++=L'*';
  *mp++=L'\0';
  hFind=FindFirstFileW(mask,&e);
  free(mask);

  if (INVALID_HANDLE_VALUE==hFind) {
#if defined(BG_TREE_CREATE_CHILD_WARNING) // [
    DVWARNINGW(L"opening directory \"%s\"",path);
#endif // ]
    goto edir;
  }
#else // ] [
  if (!(dir=opendir(path))) {
#if defined(BG_TREE_CREATE_CHILD_WARNING) // [
    DVWARNING("opening directory \"%s\"",path);
#endif // ]
    goto edir;
  }
#endif // ]

#if defined (_WIN32) // [
  do {
    name=e.cFileName;
#if 0 // [
// member dwFileType seems not to exist even if documented (cf. https://docs.microsoft.com/en-us/windows/desktop/api/minwinbase/ns-minwinbase-_win32_find_dataw)
DVWRITELN("%lu",e.dwFileType);
#endif // ]
#else // ] [
  while (NULL!=(e=readdir(dir))) {
    name=e->d_name;
#endif // ]
    if (!FFSTRCMP(FFL("."),name))
      continue;
    else if (!FFSTRCMP(FFL(".."),name))
      continue;
    else if ((BG_FLAGS_EXT_CSV&param->flags.extension)
        &&!FFSTRCMP(FFL("folder.csv"),name))
      continue;

    if (child) {
      if (bg_child_create(&child->next,param,tree->root,tree,name)<0) {
        child->next=NULL;
        continue;
      }

      child=child->next;
    }
    else {
      if (bg_child_create(&album->first,param,tree->root,tree,name)<0) {
        album->first=NULL;
        continue;
      }

      child=album->first;
    }

#if 0 && defined (FF_INPUT_ERROR_FORMAT) // [
    if (BG_TREE_TYPE_TRACK==child->vtbl->type)
      ++album->ntracks;
#endif // ]
#if ! defined (_WIN32) // [
  }
#else // ] [
  } while (FindNextFileW(hFind,&e));
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  err=0;
//cleanup:
#if defined (_WIN32) // [
  FindClose(hFind);
edir:
emask:
#else // ] [
  closedir(dir);
edir:
#endif // ]
  return err;
}

void bg_album_destroy(bg_tree_t *tree)
{
  bg_album_t *album=&tree->album;
  bg_tree_t *next;

  tree=album->first;

  while (tree) {
    next=tree->next;
    bg_child_destroy(&tree);
    free(tree);
    tree=next;
  }

  album->first=NULL;
}

///////////////////////////////////////////////////////////////////////////////
const ffchar_t *bg_album_target_purge(bg_tree_t *tree)
{
  bg_param_t *param=tree->param;
  bg_tree_t *parent=tree->parent;

  switch (tree->vtbl->type) {
  case BG_TREE_TYPE_ALBUM:
#if 0 // [
  case BG_TREE_TYPE_ROOT:
#endif // ]
    break;
  default:
    DVMESSAGE("unexpected tree type: %d",tree->vtbl->type);
    goto e_type;
  }

  // purge the tree from unnecessary basenames.
  if (!param->purged&&parent&&parent->album.first&&!parent->album.first->next)
    return NULL;

  // down the tree hierarchy stop further attempts to purge.
  param->purged=1;

  return tree->source.basename;
e_type:
  return NULL;
}

///////////////////////////////////////////////////////////////////////////////
static int bg_album_accept(bg_tree_t *tree, bg_visitor_t *vis)
{
  return vis->vtbl->dispatch_album(vis,tree);
}

static bg_tree_vtbl_t album_vtbl={
#if defined (PBU_DEBUG) // [
  .id="album",
#endif // ]
  .type=BG_TREE_TYPE_ALBUM,
  .destroy=bg_album_destroy,
  .accept=bg_album_accept,
};
